@@ -1,19 +1,21 @@
 from __future__ import absolute_import
 
-import pytest
+from datetime import timedelta
 
 import numpy as np
 import pandas as pd
-import xarray as xr
+import pytest
 
-from datetime import timedelta
+import xarray as xr
 from xarray.coding.cftimeindex import (
-    parse_iso8601, CFTimeIndex, assert_all_valid_date_type,
-    _parsed_string_to_bounds, _parse_iso8601_with_reso)
+    CFTimeIndex, _parse_array_of_cftime_strings, _parse_iso8601_with_reso,
+    _parsed_string_to_bounds, assert_all_valid_date_type, parse_iso8601)
 from xarray.tests import assert_array_equal, assert_identical
 
-from . import has_cftime, has_cftime_or_netCDF4
-from .test_coding_times import _all_cftime_date_types
+from . import (has_cftime, has_cftime_1_0_2_1, has_cftime_or_netCDF4,
+               raises_regex, requires_cftime)
+from .test_coding_times import (
+    _ALL_CALENDARS, _NON_STANDARD_CALENDARS, _all_cftime_date_types)
 
 
 def date_dict(year=None, month=None, day=None,
@@ -174,6 +176,22 @@ def test_cftimeindex_field_accessors(index, field, expected):
     assert_array_equal(result, expected)
 
 
+@pytest.mark.skipif(not has_cftime_1_0_2_1,
+                    reason='cftime not installed')
+def test_cftimeindex_dayofyear_accessor(index):
+    result = index.dayofyear
+    expected = [date.dayofyr for date in index]
+    assert_array_equal(result, expected)
+
+
+@pytest.mark.skipif(not has_cftime_1_0_2_1,
+                    reason='cftime not installed')
+def test_cftimeindex_dayofweek_accessor(index):
+    result = index.dayofweek
+    expected = [date.dayofwk for date in index]
+    assert_array_equal(result, expected)
+
+
 @pytest.mark.skipif(not has_cftime, reason='cftime not installed')
 @pytest.mark.parametrize(('string', 'date_args', 'reso'), [
     ('1999', (1999, 1, 1), 'year'),
@@ -250,16 +268,16 @@ def test_parsed_string_to_bounds_raises(date_type):
 @pytest.mark.skipif(not has_cftime, reason='cftime not installed')
 def test_get_loc(date_type, index):
     result = index.get_loc('0001')
-    expected = [0, 1]
-    assert_array_equal(result, expected)
+    assert result == slice(0, 2)
 
     result = index.get_loc(date_type(1, 2, 1))
-    expected = 1
-    assert result == expected
+    assert result == 1
 
     result = index.get_loc('0001-02-01')
-    expected = 1
-    assert result == expected
+    assert result == slice(1, 2)
+
+    with raises_regex(KeyError, '1234'):
+        index.get_loc('1234')
 
 
 @pytest.mark.skipif(not has_cftime, reason='cftime not installed')
@@ -360,7 +378,7 @@ def test_groupby(da):
 
 @pytest.mark.skipif(not has_cftime, reason='cftime not installed')
 def test_resample_error(da):
-    with pytest.raises(TypeError):
+    with pytest.raises(NotImplementedError, match='to_datetimeindex'):
         da.resample(time='Y')
 
 
@@ -594,18 +612,16 @@ def test_indexing_in_dataframe_iloc(df, index):
 
 
 @pytest.mark.skipif(not has_cftime_or_netCDF4, reason='cftime not installed')
-@pytest.mark.parametrize('enable_cftimeindex', [False, True])
-def test_concat_cftimeindex(date_type, enable_cftimeindex):
-    with xr.set_options(enable_cftimeindex=enable_cftimeindex):
-        da1 = xr.DataArray(
-            [1., 2.], coords=[[date_type(1, 1, 1), date_type(1, 2, 1)]],
-            dims=['time'])
-        da2 = xr.DataArray(
-            [3., 4.], coords=[[date_type(1, 3, 1), date_type(1, 4, 1)]],
-            dims=['time'])
-        da = xr.concat([da1, da2], dim='time')
+def test_concat_cftimeindex(date_type):
+    da1 = xr.DataArray(
+        [1., 2.], coords=[[date_type(1, 1, 1), date_type(1, 2, 1)]],
+        dims=['time'])
+    da2 = xr.DataArray(
+        [3., 4.], coords=[[date_type(1, 3, 1), date_type(1, 4, 1)]],
+        dims=['time'])
+    da = xr.concat([da1, da2], dim='time')
 
-    if enable_cftimeindex and has_cftime:
+    if has_cftime:
         assert isinstance(da.indexes['time'], CFTimeIndex)
     else:
         assert isinstance(da.indexes['time'], pd.Index)
@@ -616,3 +632,174 @@ def test_concat_cftimeindex(date_type, enable_cftimeindex):
 def test_empty_cftimeindex():
     index = CFTimeIndex([])
     assert index.date_type is None
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+def test_cftimeindex_add(index):
+    date_type = index.date_type
+    expected_dates = [date_type(1, 1, 2), date_type(1, 2, 2),
+                      date_type(2, 1, 2), date_type(2, 2, 2)]
+    expected = CFTimeIndex(expected_dates)
+    result = index + timedelta(days=1)
+    assert result.equals(expected)
+    assert isinstance(result, CFTimeIndex)
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+@pytest.mark.parametrize('calendar', _CFTIME_CALENDARS)
+def test_cftimeindex_add_timedeltaindex(calendar):
+    a = xr.cftime_range('2000', periods=5, calendar=calendar)
+    deltas = pd.TimedeltaIndex([timedelta(days=2) for _ in range(5)])
+    result = a + deltas
+    expected = a.shift(2, 'D')
+    assert result.equals(expected)
+    assert isinstance(result, CFTimeIndex)
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+def test_cftimeindex_radd(index):
+    date_type = index.date_type
+    expected_dates = [date_type(1, 1, 2), date_type(1, 2, 2),
+                      date_type(2, 1, 2), date_type(2, 2, 2)]
+    expected = CFTimeIndex(expected_dates)
+    result = timedelta(days=1) + index
+    assert result.equals(expected)
+    assert isinstance(result, CFTimeIndex)
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+@pytest.mark.parametrize('calendar', _CFTIME_CALENDARS)
+def test_timedeltaindex_add_cftimeindex(calendar):
+    a = xr.cftime_range('2000', periods=5, calendar=calendar)
+    deltas = pd.TimedeltaIndex([timedelta(days=2) for _ in range(5)])
+    result = deltas + a
+    expected = a.shift(2, 'D')
+    assert result.equals(expected)
+    assert isinstance(result, CFTimeIndex)
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+def test_cftimeindex_sub(index):
+    date_type = index.date_type
+    expected_dates = [date_type(1, 1, 2), date_type(1, 2, 2),
+                      date_type(2, 1, 2), date_type(2, 2, 2)]
+    expected = CFTimeIndex(expected_dates)
+    result = index + timedelta(days=2)
+    result = result - timedelta(days=1)
+    assert result.equals(expected)
+    assert isinstance(result, CFTimeIndex)
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+@pytest.mark.parametrize('calendar', _CFTIME_CALENDARS)
+def test_cftimeindex_sub_cftimeindex(calendar):
+    a = xr.cftime_range('2000', periods=5, calendar=calendar)
+    b = a.shift(2, 'D')
+    result = b - a
+    expected = pd.TimedeltaIndex([timedelta(days=2) for _ in range(5)])
+    assert result.equals(expected)
+    assert isinstance(result, pd.TimedeltaIndex)
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+@pytest.mark.parametrize('calendar', _CFTIME_CALENDARS)
+def test_cftimeindex_sub_timedeltaindex(calendar):
+    a = xr.cftime_range('2000', periods=5, calendar=calendar)
+    deltas = pd.TimedeltaIndex([timedelta(days=2) for _ in range(5)])
+    result = a - deltas
+    expected = a.shift(-2, 'D')
+    assert result.equals(expected)
+    assert isinstance(result, CFTimeIndex)
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+def test_cftimeindex_rsub(index):
+    with pytest.raises(TypeError):
+        timedelta(days=1) - index
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+@pytest.mark.parametrize('freq', ['D', timedelta(days=1)])
+def test_cftimeindex_shift(index, freq):
+    date_type = index.date_type
+    expected_dates = [date_type(1, 1, 3), date_type(1, 2, 3),
+                      date_type(2, 1, 3), date_type(2, 2, 3)]
+    expected = CFTimeIndex(expected_dates)
+    result = index.shift(2, freq)
+    assert result.equals(expected)
+    assert isinstance(result, CFTimeIndex)
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+def test_cftimeindex_shift_invalid_n():
+    index = xr.cftime_range('2000', periods=3)
+    with pytest.raises(TypeError):
+        index.shift('a', 'D')
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+def test_cftimeindex_shift_invalid_freq():
+    index = xr.cftime_range('2000', periods=3)
+    with pytest.raises(TypeError):
+        index.shift(1, 1)
+
+
+@requires_cftime
+def test_parse_array_of_cftime_strings():
+    from cftime import DatetimeNoLeap
+
+    strings = np.array([['2000-01-01', '2000-01-02'],
+                        ['2000-01-03', '2000-01-04']])
+    expected = np.array(
+        [[DatetimeNoLeap(2000, 1, 1), DatetimeNoLeap(2000, 1, 2)],
+         [DatetimeNoLeap(2000, 1, 3), DatetimeNoLeap(2000, 1, 4)]])
+
+    result = _parse_array_of_cftime_strings(strings, DatetimeNoLeap)
+    np.testing.assert_array_equal(result, expected)
+
+    # Test scalar array case
+    strings = np.array('2000-01-01')
+    expected = np.array(DatetimeNoLeap(2000, 1, 1))
+    result = _parse_array_of_cftime_strings(strings, DatetimeNoLeap)
+    np.testing.assert_array_equal(result, expected)
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+@pytest.mark.parametrize('calendar', _ALL_CALENDARS)
+@pytest.mark.parametrize('unsafe', [False, True])
+def test_to_datetimeindex(calendar, unsafe):
+    index = xr.cftime_range('2000', periods=5, calendar=calendar)
+    expected = pd.date_range('2000', periods=5)
+
+    if calendar in _NON_STANDARD_CALENDARS and not unsafe:
+        with pytest.warns(RuntimeWarning, match='non-standard'):
+            result = index.to_datetimeindex()
+    else:
+        result = index.to_datetimeindex(unsafe=unsafe)
+
+    assert result.equals(expected)
+    np.testing.assert_array_equal(result, expected)
+    assert isinstance(result, pd.DatetimeIndex)
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+@pytest.mark.parametrize('calendar', _ALL_CALENDARS)
+def test_to_datetimeindex_out_of_range(calendar):
+    index = xr.cftime_range('0001', periods=5, calendar=calendar)
+    with pytest.raises(ValueError, match='0001'):
+        index.to_datetimeindex()
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+@pytest.mark.parametrize('calendar', ['all_leap', '360_day'])
+def test_to_datetimeindex_feb_29(calendar):
+    index = xr.cftime_range('2001-02-28', periods=2, calendar=calendar)
+    with pytest.raises(ValueError, match='29'):
+        index.to_datetimeindex()
+
+
+@pytest.mark.skipif(not has_cftime, reason='cftime not installed')
+def test_multiindex():
+    index = xr.cftime_range('2001-01-01', periods=100, calendar='360_day')
+    mindex = pd.MultiIndex.from_arrays([index])
+    assert mindex.get_loc('2001-01') == slice(0, 30)
